#!/usr/bin/perl
#
# ===========================================================================
# confmerger.pl
# ---------------------------------------------------------------------------
# Read multiple XML Configuration files to merge them and produce
# a single one
# ---------------------------------------------------------------------------
# Author: J.Delemarche (www.trickytools.com)
# Date  : 30 Sep 2003
# Version: 0.1
# ---------------------------------------------------------------------------
# TODO: n/a
# ===========================================================================
#
use Getopt::Std;
require XML::DOM;

#
# Global Variables:
#


#
# Command line parsing:
#
getopts("v");
if ($opt_v) { $debug = 1; }

#
# Small helper functions:
#
$gDocName = "";

sub PrintError
{
	my ($errmsg) = @_;

	print STDERR "[File: $gDocName] $errmsg\n";
	print "<!-- [File: $gDocName] $errmsg -->\n";
}

sub PrintLine
{
	my ($text) = @_;

	print "<!-- $text";
	for ( $i = length($text) ; $i <= 80 - 10 ; $i++ ) {
		print " ";
	}
	print " -->\n";
}

#
# Parse the <url> content of a file:
#
%gGLOBAL = ();		# used to store global parms
%gUSERTYPE = ();	# used to store the user types
%gURL = ();		# used to store handled URL
$gCurMatch = "";	# currently analysed <url><match> tag

sub ParseGLOBAL
{
	my ($urlnode) = @_;

	$parmattr = $urlnode->getAttributes();

	$attr = $parmattr->getNamedItem('setby');
	if ($attr eq undef) { $setby = "manual"; } 
	else { $setby = $attr->getValue(); }

	$attr = $parmattr->getNamedItem('name');
	if ($attr eq undef) { $attrname = "string"; }
	else { $attrname = $attr->getValue(); }

	$attr = $parmattr->getNamedItem('value');
	if ($attr eq undef) { $value = ""; }
	else { $value = $attr->getValue(); }

	# Test if global parm already set:
	$bAlreadySet = exists($gGLOBAL{$attrname});

	if ($bAlreadySet) {
		$setby2 = $gGLOBAL{$attrname}->{'setby'};
		if ($setby2 eq '' or $setby2 eq 'manual') {
			if ($setby eq '' or $setby eq 'manual') {
				&PrintError("ERR110: Global Parameter '$attrname' is manually set more than once !");
			}

			# nothing to do:
			next;
		}

		if ($setby eq '' or $setby eq 'manual') {
			$gGLOBAL{$attrname}->{'setby'} = $setby;
			$gGLOBAL{$attrname}->{'value'} = $value;
		}
		else {
			if ($value ne $gGLOBAL{$attrname}->{'value'}) {
				&PrintError("ERR111: Global Parameter '$attrname' is automatically set more than once with different values !");
			}
		}
	}
	else {
		$gGLOBAL{$attrname}->{'setby'} = $setby;
		$gGLOBAL{$attrname}->{'value'} = $value;
	}
}

sub ParseUSERTYPE
{
	my ($urlnode) = @_;

	$attrmap  = $urlnode->getAttributes();
	$attr     = $attrmap->getNamedItem('name');
	$typename = $attr->getValue();

	# parm already set for this url ?
	$bAlreadySet = exists($gUSERTYPE{$typename});
		
	# loop on the children to get the <match> and the <parms>:
	$typechildren = $urlnode->getChildNodes();
	for ($u = 0 ; $u < $typechildren->getLength() ; $u++ ) {
		$child = $typechildren->item($u);

		if ($child->getNodeType() != XML::DOM::ELEMENT_NODE()) {
			next;
		}

		$childmap = $child->getAttributes();
		$attr     = $childmap->getNamedItem('name');
		$childname = $attr->getValue();

		if ($child->getTagName() eq "type") {
			if ($bAlreadySet) {
				if ($childname ne $gUSERTYPE{$typename}->{'type'}) {
					&PrintError("ERR200: User Type '$typename' defined more than once with different types !");
				}
			}
			else {
				$gUSERTYPE{$typename}->{'type'} = $childname;
			}

			next;
		}

		if ($child->getTagName() eq "attr") {
			$attr  = $childmap->getNamedItem('value');
			$value = $attr->getValue();

			if ($childname eq 'option') {
				$count = scalar @{$gUSERTYPE{$typename}->{'option'}};

				# option value added if not already in the array:
				$bFound = 0;
				for ( $opt = 0 ; $opt < $count ; $opt++ ) {
					if ($value eq $gUSERTYPE{$typename}->{'option'}[$opt]) {
						$bFound = 1;
					}
				}

				if ($bFound == 0) {
					$gUSERTYPE{$typename}->{'option'}[$count] = $value;
				}
				next;
			}

			# attr already set ?
			if ($bAlreadySet) {
				# auto or user set ?
				$bAttrSet = exists($gUSERTYPE{$typename}->{$childname});
				if ($bAttrSet) {
					if ($gUSERTYPE{$typename}->{$childname} ne $value) {
						&PrintError("ERR201: Attribute '$childname' of User Type '$typename' defined more than once with different values !");
					}
				}
			}
			else {
				$gUSERTYPE{$typename}->{$childname} = $value;
			}

			next;
		}
	}
}

sub ParsePARM
{
	my ($urlchild) = @_;

	$attrmap  = $urlchild->getAttributes();
	$attr     = $attrmap->getNamedItem('name');
	$parmname = $attr->getValue();

	# parm already set for this url ?
	$bAlreadySet = exists($gURL{$gCurMatch}->{$parmname});
		
	# Get the type and the children:
	$parmchildren = $urlchild->getChildNodes();
	for ( $c = 0 ; $c < $parmchildren->getLength() ; $c++ ) {
		$parmchild = $parmchildren->item($c);

		if ($parmchild->getNodeType() != XML::DOM::ELEMENT_NODE()) {
			next;
		}

		$parmattr = $parmchild->getAttributes();
		$tagname  = $parmchild->getTagName();

		$attr = $parmattr->getNamedItem('setby');
		if ($attr eq undef) { $setby = "manual"; } 
		else { $setby = $attr->getValue(); }

		$attr = $parmattr->getNamedItem('name');
		if ($attr eq undef) { $attrname = "string"; }
		else { $attrname = $attr->getValue(); }

		# 'type' child:
		if ($tagname eq 'type') {
			if ($bAlreadySet) {
				# auto or user set ?
				$setby2 = $gURL{$gCurMatch}->{$parmname}{'setby'};
				if ($setby2 eq '' or $setby2 eq 'manual') {
					if ($setby eq '' or $setby eq 'manual') {
						&PrintError("ERR100: Attribute '$attrname' of Parameter '$parmname' is manually set more than once !");
					}

					# nothing to do:
					next;
				}

				if ($setby eq '' or $setby eq 'manual') {
					$gURL{$gCurMatch}->{$parmname}{'setby'} = $setby;
					$gURL{$gCurMatch}->{$parmname}{'ptype'} = $attrname;
					next;
				}

				# check consistency:
				$ptype2 = $gURL{$gCurMatch}->{$parmname}{'ptype'};
				if ($ptype2 ne $attrname) {
					&PrintError("ERR100: Parameter '$parmname' has type of '$ptype2' which is different from former type of '$attrname'");
				}
			}
			else {
				$gURL{$gCurMatch}->{$parmname}{'setby'} = $setby;
				$gURL{$gCurMatch}->{$parmname}{'ptype'} = $attrname;
			}

			next;
		}

		# 'attr' child:
		if ($tagname eq 'attr') {
			$value = $parmattr->getNamedItem('value');
			if ($value eq undef) { $value = ""; }
			else { $value = $value->getValue(); }

			# special handling for 'option' that can be multiple:
			if ($attrname eq 'option') {
				$count = scalar @{$gURL{$gCurMatch}->{$parmname}{$attrname}->{'option'}};

				# option value added if not already in the array:
				$bFound = 0;
				for ( $opt = 0 ; $opt < $count ; $opt++ ) {
					if ($value eq $gURL{$gCurMatch}->{$parmname}{$attrname}->{'option'}[$opt]->{'value'}) {
						$bFound = 1;
					}
				}

				if ($bFound == 0) {
					$gURL{$gCurMatch}->{$parmname}{$attrname}->{'option'}[$count]->{'setby'} = $setby;
					$gURL{$gCurMatch}->{$parmname}{$attrname}->{'option'}[$count]->{'value'} = $value;
				}
				next;
			}

			# parm already set ?
			if ($bAlreadySet) {
				# auto or user set ?
				$setby2 = $gURL{$gCurMatch}->{$parmname}{$attrname}->{'setby'};
				if ($setby2 eq '' or $setby2 eq 'manual') {
					if ($setby eq '' or $setby eq 'manual') {
						&PrintError("ERR101: Attribute '$attrname' of Parameter '$parmname' is manually set more than once !");
					}

					# nothing to do:
					next;
				}

				if ($setby eq '' or $setby eq 'manual') {
					$gURL{$gCurMatch}->{$parmname}{$attrname}->{'setby'} = $setby;
					$gURL{$gCurMatch}->{$parmname}{$attrname}->{'value'} = $value;
					next;
				}

				# check consistency:
				$value2 = $gURL{$gCurMatch}->{$parmname}{$attrname}->{'value'};
				if ($value2 ne $value) {
					&PrintError("ERR102: Attribute '$attrname' of Parameter '$parmname' has different values !");
				}
			}
			else {
				$gURL{$gCurMatch}->{$parmname}{$attrname}->{'setby'} = $setby;
				$gURL{$gCurMatch}->{$parmname}{$attrname}->{'value'} = $value;
			}

			next;
		}
	}
}

sub ParseURL
{
	my ($urlnode) = @_;

	# loop on the children to get the <match> and the <parms>:
	$urlchildren = $urlnode->getChildNodes();
	for ($u = 0 ; $u < $urlchildren->getLength() ; $u++ ) {
		$urlchild = $urlchildren->item($u);

		if ($urlchild->getNodeType() != XML::DOM::ELEMENT_NODE()) {
			next;
		}

		if ($urlchild->getTagName() eq "match") {
			# get the <match> text note:
			$pattern = "";
			$matchchildren = $urlchild->getChildNodes();
			for ( $m = 0 ; $m < $matchchildren->getLength() ; $m++ )  {
				$matchchild = $matchchildren->item($m);
				if ($matchchild->getNodeType() == XML::DOM::TEXT_NODE()) {
					$pattern = $matchchild->getData();
					last;
				}
			}
			if ($pattern eq "") {
				&PrintError("WARN100: <match> tag without any text node !");
				next;
			}

			# TODO: display a warning when URL are similar (ignore the leading caret ?)
			if (!exists($gURL{$pattern})) {
				$gURL{$pattern} = ();
			}
			$gCurMatch = $pattern;
			next;
		}

		if ($urlchild->getTagName() eq "parm") {
			&ParsePARM($urlchild);
			next;
		}
	}
}

#
# "Main" function here:
#
$parser  = new XML::DOM::Parser;

# Parse the given document and build their DOMs:
@docs = ();
@docname = ();
for ( $a = 0 ; $a < @ARGV ; $a++ ) {
	if ($ARGV[$a] eq "-v") { next; }

	$doc = $parser->parsefile($ARGV[$a]);
	push @docs,$doc;
	push @docname,$ARGV[$a];
}

if (@docs == 0) {
	print STDERR "Usage: $0 [-v]\n";
	print STDERR "\t-v: verbose (debug) mode\n";
	exit 1;
}

# Hard stuff here !
# Loop on all url described in all the docs...
# Note: in this version, comments are not saved...

for ( $d = 0 ; $d < @docs ; $d++ ) {
	$doc = $docs[$d];
	$gDocName = $docname[$d];
	if ($debug) { print STDERR "Analysis of file $gDocName...\n"; }
	
	# handle the global parms:
	$nodelist = $doc->getElementsByTagName("global");
	for ( $n = 0 ; $n < $nodelist->getLength() ; $n++ ) {
		$urlnode = $nodelist->item($n);
		&ParseGLOBAL($urlnode);
	}

	# handle the user types:
	$nodelist = $doc->getElementsByTagName("usertype");
	for ( $n = 0 ; $n < $nodelist->getLength() ; $n++ ) {
		$urlnode = $nodelist->item($n);
		&ParseUSERTYPE($urlnode);
	}


	# loop on the <url>:
	$nodelist = $doc->getElementsByTagName("url");
	for ( $n = 0 ; $n < $nodelist->getLength() ; $n++ ) {
		$urlnode = $nodelist->item($n);
		&ParseURL($urlnode);
	}
}

# 
# Output the results now !
#
print "<?xml version=\"1.0\">\n";
print "<!DOCTYPE parmguard SYSTEM \"mod_parmguard.dtd\">\n";
print "\n";
print "<!-- ======================================================================= -->\n";
&PrintLine("CONFIGURATION MERGER");
&PrintLine("mod_parmguard Generator, version 1.2");
$now = localtime;
&PrintLine("Date of Merge: $now"); 
&PrintLine("List of merged files:"); 
for ( $a = 0 ; $a < @ARGV ; $a++ ) {
	&PrintLine($ARGV[$a]);
}
print "<!-- ======================================================================= -->\n";
print "\n";
print "<parmguard>\n";

foreach $key (keys %gGLOBAL) {
	print "\t<global setby=\"";
	if (!exists($gGLOBAL{$key}->{'setby'})) { print "auto"; }
	else { print $gGLOBAL{$key}->{'setby'}; }
	print "\" name=\"$key\" value=\"";
	print $gGLOBAL{$key}->{'value'};
	print "\"/>\n";
}
print "\n";

foreach $key (keys %gUSERTYPE) {
	print "\t<usertype name=\"$key\">\n";;
	print "\t\t<type name=\"",$gUSERTYPE{$key}->{'type'},"\"/>\n";

	foreach $subkey (keys %{$gUSERTYPE{$key}}) {
		if ($subkey eq 'type') { next; }
		if ($subkey eq 'option') {
			for ( $opt = 0 ; $opt < @{$gUSERTYPE{$key}->{'option'}} ; $opt++ ) {
				print "\t\t<attr name=\"option\" value=\"",$gUSERTYPE{$key}->{'option'}[$opt],"\"/>\n";
			}
			next;
		}

		print "\t\t<attr name=\"$subkey\" value=\"",$gUSERTYPE{$key}->{$subkey},"\"/>\n";
	}

	print "\t</usertype>\n";
}
print "\n";

foreach $key (keys %gURL) {
	print "\t<url>\n";
	print "\t\t<match>$key</match>\n";

	foreach $subkey (keys %{$gURL{$key}}) {
		if ($subkey eq 'global') {
			next;
		}

		print "\t\t<parm name=\"$subkey\">\n";

		print "\t\t\t<type setby=\"";
		if (!exists($gURL{$key}->{$subkey}{'setby'})) { print "auto"; }
		else { print $gURL{$key}->{$subkey}{'setby'}; }
		print "\" name=\"";
		if (!exists($gURL{$key}->{$subkey}{'ptype'})) { print "string"; }
		else { print $gURL{$key}->{$subkey}{'ptype'}; }
		print "\"/>\n";

		for $attr (keys %{$gURL{$key}{$subkey}}) {
			if ($attr eq 'ptype' or $attr eq 'setby') { next; }
			if ($attr ne 'option') {
				print "\t\t\t<attr setby=\"";
				if (!exists($gURL{$key}->{$subkey}{$attr}->{'setby'})) { print "auto"; }
				else { print $gURL{$key}->{$subkey}{$attr}->{'setby'}; }
				print "\" name=\"$attr\" value=\"";
				if (!exists($gURL{$key}->{$subkey}{$attr}->{'value'})) { print ""; }
				else { print $gURL{$key}->{$subkey}{$attr}->{'value'}; }
				print "\"/>\n";
				next;
			}

			for ( $opt = 0 ; $opt < @{$gURL{$key}->{$subkey}{$attr}->{'option'}} ; $opt++ ) {
				print "\t\t\t<attr setby=\"";
				if (!exists($gURL{$key}->{$subkey}{$attr}->{'option'}[$opt]->{'setby'})) { print "auto"; }
				else { print $gURL{$key}->{$subkey}{$attr}->{'option'}[$opt]->{'setby'}; }
				print "\" name=\"$attr\" value=\"";
				if (!exists($gURL{$key}->{$subkey}{$attr}->{'option'}[$opt]->{'value'})) { print ""; }
				else { print $gURL{$key}->{$subkey}{$attr}->{'option'}[$opt]->{'value'}; }
				print "\"/>\n";
			}
		}

		print "\t\t</parm>\n";
	}

	print "\t</url>\n";
}

print "</parmguard>\n";

exit 0;

# EOF
