#!/usr/bin/perl
#
# ===========================================================================
# htmlspider.pl
# ---------------------------------------------------------------------------
# Recursively scan the given URL and all the related links to build
# a XML Configuration file for mod_parmguard.
# ---------------------------------------------------------------------------
# Author: J.Delemarche (www.trickytools.com)
# Date  : 30 Sep 2003
# Version: 0.1
# ---------------------------------------------------------------------------
# TODO: n/a
# ===========================================================================
#
use Getopt::Std;
use HTML::Parser;
use HTTP::Response;
use URI::URL;
use File::Basename;


#
# Global Variables:
#
$gHomeURL   = "";
$gAuthority = "";
$gBaseDir   = "";
$gBaseTag   = "";
$debug      = 0;
$gUserAgent = "mod_parmguard/1.2";

@todoHREF = ();		# URLs to be parsed
%doneHREF = ();		# attributes of Parsed URLs
%codeHREF = ();		# HTTP error code of parsed URLs

#
# Command line parsing:
#
getopts("vu:h:");
if ($opt_v) { $debug = 1; }
if ($opt_u) { $gUserAgent = $opt_u; }
if ($opt_h) { $gHomeURL = $opt_h; }

# Help or command line error ?
if ($gHomeURL eq "") {
	print STDERR "Usage: $0 [-v] [-u useragent] -h homeURL\n";
	print STDERR "\t-v: verbose (debug) mode\n";
	print STDERR "\t-u: useragent (what will appear in the server log)\n";
	print STDERR "\t-h: homeURL (start point of the scanning)\n";
	exit 1;
}

# 
# My LWP::UserAgent class to avoid redirect:
#
package MyLWPUserAgent;
use LWP::UserAgent;
@ISA = ("LWP::UserAgent");

sub redirect_ok
{
	return 0;
}

# 
# Class to handle the parsed URL:
#
package MyURL;

sub new
{
	my $class = shift;
	my $self  = {};

	bless($self,$class);

	return $self;
}

#
# Go back to main package:
#
package main;

#
# Small helper functions:
#
sub PrintError
{
	my ($errmsg) = @_;

	print STDERR "$errmsg\n";
	print "<!-- $errmsg -->\n";
}

sub PrintLine
{
	my ($text) = @_;

	print "<!-- $text";
	for ( $i = length($text) ; $i <= 80 - 10 ; $i++ ) {
		print " ";
	}
	print " -->\n";
}

sub SetBase
{
	my ($url) = @_;
	my $uri = URI->new($url);
	my $idx = rindex($uri->path,"/");
	if ($idx == -1) {
		$gBaseDir = "/";
	}
	else {
		$gBaseDir = substr($uri->path,0,$idx+1);
	}
		
	if ($debug) {
		print STDERR "New gBaseDir=$gBaseDir (from $url, path=",$uri->path,")\n";
	}
}

sub BuildURL
{
	my ($url) = @_;
	my $newurl,$uri;

	#print STDERR "\tPushURL: $url\n";

	# assume we stay in the same domain:
	$uri = URI->new($url);
	$baseuri = URI->new($gBaseTag);

	# scheme must be http:
	if ($uri->scheme ne "" && $uri->scheme ne "http") {
		return "";
	}

	# the authority must be under the root authority:
	if ($uri->authority ne "" && $uri->authority ne $gAuthority) {
		return "";
	}
	if ($uri->authority eq "") {
		if ($baseuri->authority ne "" && $baseuri->authority ne $gAuthority) {
			return "";
		}
	}

	# Canonicalize:
	$file = $uri->path;
	while (substr($file,0,2) eq "./") {
		$file = substr($file,2);
	}

	if ($baseuri->path ne "") {
		$base = $baseuri->path;
	}
	else {
		$base = $gBaseDir;
	}
	while (substr($file,0,3) eq "../") {
		if ($base ne "/") {
			$base = dirname($base)."/";
		}
		$file = substr($file,3);
	}

	# Add scheme:
	$newurl = "http://"; 

	# gAuthority:
	$newurl .= $uri->authority;
	if ($uri->authority eq "") { 
		$newurl .= $gAuthority; 
		if (substr($uri->path,0,1) ne "/") { 
			$newurl .= $base; 
		}
	}

	# Path:
	if ($file eq "" or $file eq "/") { return ""; }
	$newurl .= $file;

	# Query: (ignore the fragment)
	if ($uri->query) {
		$newurl .= "?".$uri->query;
	}

	$uri = URI->new($newurl);
	$newurl = $uri->canonical();

	return $newurl;
}

#
# Handling of URL and result parsing:
#
%gURL = ();

#
# Callback for HTML Parser
#
$gCurURL 	= "";
$gFormAction 	= "";
$gFormMethod 	= "";
$gSelectOption 	= "";

sub AddURL
{
	my ($url) = @_;

	if (!exists($gURL{$url})) {
		#print "First Time URL \n";
		$gURL{$url}->{'url'} = 1;
	}
	else {
		#print "NOT First Time URL \n";
	}
}

sub PushURL
{
	my ($url,$method) = @_;

	$newurl = &BuildURL($url,$method);
	if ($newurl eq "") { return; }

	if (!exists $doneHREF{$newurl}) {
		if ($debug) {
			print STDERR "\tPushURL: $newurl\n";
		}

		push @todoHREF,$newurl;
		$doneHREF{$newurl} = $method;
	}

	# Check if the url contains a query or not:
	$uri = URI->new($newurl);
	if ($uri->query ne "") {
		@qf = $uri->query_form;
		&AddURL($uri->path);

		for ( $i = 0 ; $i < @qf ; $i += 2 ) {
			$gURL{$uri->path}->{$qf[$i]}{'type'}  = 'text';
		}
	}
	
}

sub StartForm
{
	my ($url) = @_;

	$gFormAction = "";
	$newurl = BuildURL($url);
	if ($newurl eq "") { return; }

	$uri = URI->new($newurl);

	$gFormAction = $uri->path;
	&AddURL($gFormAction);
}

sub cback_start
{
	my ($tag,$attr) = @_;

	#print STDERR "$tag\n";
	if ($tag eq "base") {
		$gBaseTag = $attr->{'href'};
	}
	elsif ($tag eq "meta") {
		if ($attr->{'http-equiv'} eq "refresh") {
			$value = $attr->{'content'};
			$value =~ /url=(.*);*/;
			$href = $1;
			$href =~ s/;//;
			&PushURL($href,'GET');
		}
	}
	elsif ($tag eq "applet") {
		&PushURL($attr->{'codebase'},'GET');
	}
	elsif ($tag eq "a" || $tag eq "area" || $tag eq "link") {
		&PushURL($attr->{'href'},'GET');
	}
	elsif ($tag eq "table" || $tag eq "body") {
		&PushURL($attr->{'background'},'GET');
	}
	elsif ($tag eq "head") {
		&PushURL($attr->{'profile'},'GET');
	}
	elsif ($tag eq "frame" || $tag eq "iframe") {
		&PushURL($attr->{'src'},'GET');
		&PushURL($attr->{'longdesc'},'GET');
	}
	elsif ($tag eq "img") {
		&PushURL($attr->{'src'},'GET');
		&PushURL($attr->{'src'},'GET');
		&PushURL($attr->{'src'},'GET');
		&PushURL($attr->{'dynsrc'},'GET');
		&PushURL($attr->{'usemap'},'GET');
	}
	elsif ($tag eq "script" || $tag eq "bgsound" || $tag eq "embed" || $tag eq "ilayer") {
		&PushURL($attr->{'src'},'GET');
	}
	elsif ($tag eq "ins" || $tag eq "q") {
		&PushURL($attr->{'cite'},'GET');
	}
	elsif ($tag eq "isindex") {
		&PushURL($attr->{'action'},'GET');
	}
	elsif ($tag eq "form") {
		if ($attr->{'method'} ne "") { $gFormMethod = $attr->{'method'}; }
		else { $gFormMethod = "GET"; }
		&StartForm($attr->{'action'});
		&PushURL($attr->{'action'},uc($gFormMethod));

		if ($attr->{'action'} eq "") {
			&PrintError("WARN003: [$gCurURL] <form> tag without any action in form $gFormAction");
		}
	}
	elsif ($tag eq "input") {
		#print "input for form $gFormAction\n";
		if ($gFormAction ne "") {
			if ($attr->{'name'} eq "") {
				&PrintError("WARN001: [$gCurURL] <input> tag without a name in form $gFormAction");
			}
			else {
				$type = $attr->{'type'};
				$name = $attr->{'name'};
				$value = $attr->{'value'};

				$gURL{$gFormAction}->{$name}{'type'}  = $type;
				$gURL{$gFormAction}->{$name}{'value'} = $value;
				if (exists $attr->{'maxlength'}) {
					$gURL{$gFormAction}->{$name}{'maxlength'} = $attr->{'maxlength'};
				}

				# BUTTON:
				if ($type eq "button") {
				}
				# CHECKBOX:
				elsif ($type eq "checkbox") {
				}
				# IMAGE:
				elsif ($type eq "image") {
				}
				# TEXTEAREA:
				elsif ($type eq "textarea") {
				}
				# HIDDEN & PASSWORD & TEXT & SUBMIT:
				elsif ($type eq "hidden" || $type eq "password" || $type eq "text" || $type eq "submit") {
				}
				# RADIO:
				elsif ($type eq "radio") {
					# Memorize the radio buttons of same groups: 
					# Push only if not already in the list:
					$found = 0;
					for $v (@{$gURL{$gFormAction}->{$name}{'radio'}}) {
						if ($v eq $value) {
							$found = 1;
						}
					}
					if ($found != 1) {
						push @{$gURL{$gFormAction}->{$name}{'radio'}},$value;
					}
				}
			}
		}
	}
	elsif ($tag eq "select") {
		if ($gFormAction ne "") {
			if ($attr->{'name'} eq "") {
				&PrintError("WARN004: [$gCurURL] <select> tag without a name in form $gFormAction");
			}
			else {
				$name = $attr->{'name'};
				$gURL{$gFormAction}->{$name}{'type'}  = 'select';

				$gSelectOption = $name;
				$multiple = 0;
				if (exists($attr->{'multiple'})) {
					$multiple = 1;
				}
				$gURL{$gFormAction}->{$name}{'multiple'}  = $multiple;

			}
		}
	}
	elsif ($tag eq "option") {
		if ($gSelectOption ne "") {
			$value = $attr->{'value'};
			push @{$gURL{$gFormAction}->{$gSelectOption}{'option'}},$value;
		}
	}
	elsif ($tag eq "layer") {
		&PushURL($attr->{'src'},'GET');
		&PushURL($attr->{'background'},'GET');
	}
	elsif ($tag eq "object") {
		&PushURL($attr->{'data'},'GET');
		&PushURL($attr->{'usemap'},'GET');
	}
}

sub cback_end
{
	my ($tag) = @_;

	if ($tag eq "form") {
		$gFormAction = "";
	}
}

#
# Real work starts here:
#
sub RequestCallback
{
	my ($chunk) = @_;
	#print STDERR $chunk;

	$parser->parse($chunk);
}

sub ScanURL
{
	# the URL is supposed to be canonicalized
	my ($url) = @_;
	my $method = $doneHREF{$url};

	if ($debug) { print STDERR ">>>ScanURL: $method $url\n"; }

	# The new url becomes the Base for the included links:
	$gCurURL = $url;
	&SetBase($url);
	$gBaseTag = "";

	$request = HTTP::Request->new($method,$url);
	$doc = $ua->request($request,\&RequestCallback);

	if ($debug) {
		print STDERR "<<<ScanUrl: $method $url: ";
		if ($doc->is_redirect()) { print STDERR "[REDIRECT: ",$doc->header('Location'),"]\n"; }
		if ($doc->is_error()) { print STDERR "[FAILED=",$doc->code(),"]\n"; }
		if ($doc->is_success()) { print STDERR "[OK]\n"; }
	}
	if ($doc->is_error()) {
		$codeHREF{$url} = $doc->code();
	}

	if ($doc->is_redirect()) {
		$newloc = $doc->header('Location');
		&PushURL($newloc,$method);
	}
}

#
# "Main" function here:
#
$parser = HTML::Parser->new
			(	api_version => 3,
				start_h     => [\&cback_start, "tagname, attr" ],
				end_h       => [\&cback_end,   "tagname, attr" ],
			);
$ua     = MyLWPUserAgent->new;
$ua->agent($gUserAgent);

#
# Parsing:
#
$uri = URI->new($gHomeURL);
$gAuthority = $uri->authority;
&SetBase($gHomeURL);

push @todoHREF,$gHomeURL;
$doneHREF{$gHomeURL} = 'GET';
while (@todoHREF > 0) {
	$url = pop @todoHREF;
	&ScanURL($url);
}

# 
# Output the results now !
#
print "<?xml version=\"1.0\"?>\n";
print "<!DOCTYPE parmguard SYSTEM \"mod_parmguard.dtd\">\n";
print "\n";
print "<!-- ======================================================================= -->\n";
&PrintLine("SCANNING SUMMARY");
&PrintLine("mod_parmguard Generator, version 1.2");
$now = localtime;
&PrintLine("Date of Scan: $now"); 
&PrintLine("Start URL: $gHomeURL");
&PrintLine("List of not parsed URLs"); 
if (scalar keys %codeHREF == 0) {
	&PrintLine("(none)");
}
else {
	while (($key,$value) = each(%codeHREF)) {
		&PrintLine("$key: $value"); 
	}
}
print "<!-- ======================================================================= -->\n";
print "\n";
print "<parmguard>\n";

foreach $key (keys %gURL) {
	print "\t<url>\n";
	print "\t\t<match>^$key</match>\n";

	if (defined $gURL{$key}) {
		foreach $input (keys %{$gURL{$key}}) {
			$value = $gURL{$key}->{$input}{'value'};
			$type  = $gURL{$key}->{$input}{'type'};

			if ($type eq 'button') {
				print "\t\t<parm name=\"$input\">\n";
				print "\t\t\t<type setby=\"auto\" name=\"enum\"/>\n";
				print "\t\t\t<attr setby=\"auto\" name=\"option\" value=\"$value\"/>\n";
				print "\t\t</parm>\n";
			}
			elsif ($type eq 'checkbox') {
				print "\t\t<parm name=\"$input\">\n";
				print "\t\t\t<type setby=\"auto\" name=\"enum\"/>\n";
				print "\t\t\t<attr setby=\"auto\" name=\"option\" value=\"$value\"/>\n";
				print "\t\t\t<attr setby=\"auto\" name=\"option\" value=\"\"/>\n";
				print "\t\t</parm>\n";
			}
			elsif ($type eq 'image') {
				print "\t\t<parm name=\"$input.x\">\n";
				print "\t\t\t<type setby=\"auto\" name=\"integer\"/>\n";
				print "\t\t</parm>\n";
				print "\t\t<parm name=\"$input.y\">\n";
				print "\t\t\t<type setby=\"auto\" name=\"integer\"/>\n";
				print "\t\t</parm>\n";
			}
			elsif ($type eq 'textarea') {
				print "\t\t<parm name=\"$input\">\n";
				print "\t\t\t<type setby=\"auto\" name=\"string\"/>\n";
				print "\t\t</parm>\n";
			}
			elsif ($type eq 'hidden' || $type eq 'text' || $type eq 'password' || $type eq 'submit') {
				print "\t\t<parm name=\"$input\">\n";

				# Use the 'class' attribute as the type:
				if (exists $gURL{$key}->{$input}{'class'}) {
					$class = $gURL{$key}->{$input}{'class'};
					print "\t\t\t<type setby=\"auto\" name=\"$class\"/>\n";

					if ($class eq "") {
						&PrintError("WARN002: [$gCurURL] <input> tag without a class name in form $key");
					}
				}
				else {
					print "\t\t\t<type setby=\"auto\" name=\"string\"/>\n";
				}

				# Use the 'maxsize' attribute:
				if ($type eq "password" || $type eq "text") {
					if (exists $gURL{$key}->{$input}{'maxlength'}) {
						$maxlen = $gURL{$key}->{$input}{'maxlength'};
						print "\t\t\t<attr setby=\"auto\" name=\"maxlen\" value=\"$maxlen\"/>\n";
					}
				}

				print "\t\t</parm>\n";
			}
			elsif ($type eq 'radio') {
				print "\t\t<parm name=\"",$input,"\">\n";
				print "\t\t\t<type setby=\"auto\" name=\"enum\"/>\n";
				print "\t\t\t<attr setby=\"auto\" name=\"multiple\" value=\"0\"/>\n";

				for $v (@{$gURL{$key}->{$input}{'radio'}}) {
					print "\t\t\t<attr setby=\"auto\" name=\"option\" value=\"$v\"/>\n";
				}

				print "\t\t</parm>\n";
			}
			elsif ($type eq 'select') {
				$multiple = $gURL{$key}->{$input}{'multiple'};
				print "\t\t<parm name=\"$input\">\n";
				print "\t\t\t<type setby=\"auto\" name=\"enum\"/>\n";
				print "\t\t\t<attr setby=\"auto\" name=\"multiple\" value=\"$multiple\"/>\n";

				for $v (@{$gURL{$key}->{$input}{'option'}}) {
					print "\t\t\t<attr setby=\"auto\" name=\"option\" value=\"$v\"/>\n";
				}

				print "\t\t</parm>\n";
			}
		}
	}

	print "\t</url>\n";
}

print "</parmguard>\n";

exit 0;

# EOF
